//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "PseudoEvent.hpp"

#include "MyException.hpp"
#include "Screen.hpp"
#include "Global.hpp"
#include "nbStrConvert.hpp"

#include <iostream>

#include <lua.hpp>
#include <luabind/luabind.hpp>
//------------------------------------------------------------------------------
namespace En
{
using namespace std;
//------------------------------------------------------------------------------
PseudoEventHandler::PseudoEventHandler(const luabind::object &luaFunction) : EventHandlerBase(luaFunction)
{

}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Timer::Timer(const luabind::object &luaFunction, float interval, bool isRunning) : PseudoEventHandler(luaFunction), m_currentTime(0), m_interval(interval), m_isRunning(isRunning)
{

}
//------------------------------------------------------------------------------
bool Timer::Match(const Screen *screen, const sf::RenderWindow &App)
{
    bool match = false;

    if(IsRunning())
    {
        m_currentTime += App.GetFrameTime() * screen->GetTimeSpeed();

        size_t nb = m_currentTime / m_interval;
        for(size_t i = 0; i < nb; i++)
        {
            SendEvent(this, m_luaFunction);
            match = true;

            if(!IsRunning() || m_isEndOfLife)
            {
                break;
            }
        }

        m_currentTime -= nb * m_interval;
    }

    return match;
}
//------------------------------------------------------------------------------
void Timer::SendEvent(Timer *obj, const luabind::object &luaFunction)
{
    if(! luaFunction.is_valid())
    {
        return;
    }

    try
    {
        luabind::call_function<void>(luaFunction, ProtectedPtr<Timer>(this));
    }
    MY_CATCH_LUA_CERR()
}
//------------------------------------------------------------------------------
/** Get the timer elapsed time for the current tick. */
float Timer::GetElapsedTime() const
{
    return m_currentTime;
}
//------------------------------------------------------------------------------
/** Set the timer elapsed time for the current tick. */
void Timer::SetElapsedTime(float elapsedTime)
{
    elapsedTime = max(elapsedTime, 0.f);
    if(elapsedTime > m_interval)
    {
        MY_THROW("Timer::SetElapsedTime() : elapsedTime > interval !")
    }
    else
    {
        m_currentTime = elapsedTime;
    }
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Chronometer::Chronometer(const luabind::object &luaFunction, float totalTime, bool isRunning) : PseudoEventHandler(luaFunction), m_currentTime(0), m_totalTime(totalTime), m_isRunning(isRunning)
{

}
//------------------------------------------------------------------------------
bool Chronometer::Match(const Screen *screen, const sf::RenderWindow &App)
{
    if(IsRunning())
    {
        m_currentTime += App.GetFrameTime() * screen->GetTimeSpeed();
        if(m_currentTime >= m_totalTime)
        {
            m_currentTime = m_totalTime;
            Stop();
        }

        SendEvent(GetRatio(), this, m_luaFunction);

        return true;
    }
    else
    {
        return false;
    }
}
//------------------------------------------------------------------------------
/** Force the chronometer to force its current ratio */
void Chronometer::SetRatio(float newRatio)
{
    m_currentTime = min(max(newRatio, 0.f), 1.f) * m_totalTime;
}
//------------------------------------------------------------------------------
void Chronometer::SendEvent(float ratio, Chronometer *obj, const luabind::object &luaFunction)
{
    if(! luaFunction.is_valid())
    {
        return;
    }

    try
    {
        luabind::call_function<void>(luaFunction, ratio, ProtectedPtr<Chronometer>(obj));
    }
    MY_CATCH_LUA_CERR()
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
FrameEvent::FrameEvent(const luabind::object &luaFunction) : PseudoEventHandler(luaFunction)
{

}
//------------------------------------------------------------------------------
bool FrameEvent::Match(const Screen *screen, const sf::RenderWindow &App)
{
    SendEvent(App.GetFrameTime() * screen->GetTimeSpeed(), this, m_luaFunction);
    return true;
}
//------------------------------------------------------------------------------
void FrameEvent::SendEvent(float frameTime, FrameEvent *obj, const luabind::object &luaFunction)
{
    if(! luaFunction.is_valid())
    {
        return;
    }

    try
    {
        luabind::call_function<void>(luaFunction, frameTime, ProtectedPtr<EventHandlerBase>(obj));
    }
    MY_CATCH_LUA_CERR()
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

OneCallbackEvent::OneCallbackEvent(const luabind::object &luaFunction, std::vector<luabind::object> argList) : PseudoEventHandler(luaFunction), m_argList(argList)
{

}
//------------------------------------------------------------------------------
bool OneCallbackEvent::Match(const Screen *screen, const sf::RenderWindow &App)
{
    SendEvent(m_argList, m_luaFunction);
    Destroy(); // Only one call
    return true;
}
//------------------------------------------------------------------------------
void OneCallbackEvent::SendEvent(const std::vector<luabind::object> &argList, const luabind::object &luaFunction)
{
    if(! luaFunction.is_valid())
    {
        return;
    }

    switch(argList.size())
    {
    case 0:
        try
        {
            luabind::call_function<void>(luaFunction);
        }
        MY_CATCH_LUA_CERR()
        break;
    case 1:
        try
        {
            luabind::call_function<void>(luaFunction, argList[0]);
        }
        MY_CATCH_LUA_CERR()
        break;
    case 2:
        try
        {
            luabind::call_function<void>(luaFunction, argList[0], argList[1]);
        }
        MY_CATCH_LUA_CERR()
        break;
    default:
        MY_THROW("OneCallbackEvent::SendEvent() : too many args !!!");
        break;
    }
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
MouseZoneEvent::MouseZoneEvent(const luabind::object &luaFunction, ProtectedPtr<PhysicObject> zone) : PseudoEventHandler(luaFunction), m_zone(zone), m_isIn(false)
{

}
//------------------------------------------------------------------------------
bool MouseZoneEvent::Match(const Screen *screen, const sf::RenderWindow &App)
{
    if(!m_zone.IsValid())
    {
        MY_THROW("MouseZoneEvent: The given PhysicObject (zone) have been destroyed !")
    }

    bool status = m_zone->IsPointInside(Global::GetMousePosition());

    if(status != m_isIn)
    {
        m_isIn = status;
        SendEvent(this, m_luaFunction);
        return true;
    }
    else
    {
        return  false;
    }
}
//------------------------------------------------------------------------------
void MouseZoneEvent::SendEvent(MouseZoneEvent *obj, const luabind::object &luaFunction)
{
    if(! luaFunction.is_valid())
    {
        return;
    }

    try
    {
        luabind::call_function<void>(luaFunction, ProtectedPtr<MouseZoneEvent>(obj));
    }
    MY_CATCH_LUA_CERR()

}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
