//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef BOTTLE_UTILS_HPP
#define BOTTLE_UTILS_HPP
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
#include <SFML/Audio.hpp>

#include "ClassAllocCounter.hpp"

#include "ProtectedPtr.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
class Screen;
class SoundPlayer;
//------------------------------------------------------------------------------
/* Vector3D */
/// LUA_HEADER_COPY_ON ///
/** A simple vector in three dimentions. */
/// LUA_HEADER_COPY_OFF ///

class Vector3D : public sf::Vector3f
{
public:
    // Depuis df::Vector3f
    Vector3D(const sf::Vector3f &vect) : sf::Vector3f(vect) {}

    /// LUA_EXPORT_ON(Vector3D) ///
    // Constructor
    Vector3D(float x, float y, float z) : sf::Vector3f(x, y, z) {}
    Vector3D() : sf::Vector3f() {}

    // Operators
    /** Add two vectors */
    Vector3D operator +(const Vector3D& vect) const;
    /** Subtract two vectors */
    Vector3D operator -(const Vector3D& vect) const;
    /** Check equality between two vectors */
    bool operator ==(const Vector3D& vect) const;
    /** Multiply a Vector by a number */
    Vector3D operator *(float nb) const;
    /** Divide a Vector by a number */
    Vector3D operator /(float nb) const;

    // Special operations
    /** Get the sum of the squares of x and y (basicaly, x² + y² + z²). This is useful to compare the vectors size without using a square root (which is more time consuming). */
    float GetSquareSum() const;

    /** Get the size of the Vector */
    float GetSize() const;

    /** Get the normalised vector (direction). */
    Vector3D GetDirection() const;

    /** Get the dot product / scalar product of this vector with another one */
    float GetDotProduct(const Vector3D &vector2) const;

    // Acces
    /** Access the x componant */
    //LUA// luaReadwrite(float, x);
    /** Access the y componant */
    //LUA// luaReadwrite(float, y);
    /** Access the z componant */
    //LUA// luaReadwrite(float, z);
    /// LUA_EXPORT_OFF ///
};
//------------------------------------------------------------------------------
/* Vector2D */
/// LUA_HEADER_COPY_ON ///
/** A simple vector in two dimentions. */
/// LUA_HEADER_COPY_OFF ///

class Vector2D : public sf::Vector2f
{
public:
    // Depuis df::Vector2f
    Vector2D(const sf::Vector2f &vect) : sf::Vector2f(vect) {}

    /// LUA_EXPORT_ON(Vector2D) ///
    // Constructor
    Vector2D(float x, float y) : sf::Vector2f(x, y) {}
    Vector2D() : sf::Vector2f() {}

    // Operators
    /** Add two vectors */
    Vector2D operator +(const Vector2D& vect) const;
    /** Subtract two vectors */
    Vector2D operator -(const Vector2D& vect) const;
    /** Check equality between two vectors */
    bool operator ==(const Vector2D& vect) const;
    /** Multiply a Vector by a number */
    Vector2D operator *(float nb) const;
    /** Divide a Vector by a number */
    Vector2D operator /(float nb) const;

    // Special operations
    /** Get the sum of the squares of x and y (basicaly, x² + y²). This is useful to compare the vectors size without using a square root (which is more time consuming). */
    float GetSquareSum() const;

    /** Get the size of the Vector */
    float GetSize() const;

    /** Get the normalised vector (direction). */
    Vector2D GetDirection() const;

    /** Get the dot product / scalar product of this vector with another one */
    float GetDotProduct(const Vector2D &vector2) const;

    /** Get the angle between this vector and the x axis */
    float GetAngle() const;

    /** Get the normal to the vector, in the <b>direct</b> direction (<b>+</b> pi/2). The returned coordonates are Vector2D(-y, x). */
    Vector2D GetDirectNormal() const;

    /** Get the normal to the vector, in the <b>indirect</b> direction (<b>-</b> pi/2). The returned coordonates are Vector2D(y, -x). */
    Vector2D GetIndirectNormal() const;

    // /** Get the angle between the given vector (reference) and this vector */
    //float GetAngle(const Vector2D &vector2) const;

    // Acces
    /** Access the x componant */
    //LUA// luaReadwrite(float, x);
    /** Access the y componant */
    //LUA// luaReadwrite(float, y);
    /// LUA_EXPORT_OFF ///
};
//------------------------------------------------------------------------------
sf::Unicode::UTF8String StrToSfUtf8(const std::string &str);
Vector2D PointRotate(const Vector2D &centre, const Vector2D &pt, float angle);
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

typedef sf::Color Color;
/// LUA_HEADER_COPY_ON ///
/** A color, with red, green, blue and alpha value. Min is 0 and max is 255. */
/// LUA_HEADER_COPY_OFF ///

/// LUA_EXPORT_ON(Color) ///
//LUA// Color();
//LUA// Color(char r, char g, char b);
//LUA// Color(char r, char g, char b, char a);
//LUA// luaReadwrite(char, r);
//LUA// luaReadwrite(char, g);
//LUA// luaReadwrite(char, b);
//LUA// luaReadwrite(char, a);
/// LUA_EXPORT_OFF ///
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A texture (an image) that can be loaded from a file. */
/// LUA_HEADER_COPY_OFF ///

class Texture : public sf::Image, public ClassAllocCounter, public ProtectedClassBase
{
public:
    /// LUA_EXPORT_ON(Texture) ///
    /** Constructor to create an empty texture */
    Texture();

    /** Create a texture from a file */
    Texture(const std::string &filename);

    /** Return the width of the image. */
    float GetWidth() const {return sf::Image::GetWidth();}

    /** Return the height of the image. */
    float GetHeight() const {return sf::Image::GetHeight();}

    /** Return the size of the image. */
    Vector2D GetSize() const {return Vector2D(GetWidth(), GetHeight());}

    /** Load a texture from a file. Note: the working directory is the mode's root directory. */
    void LoadFromFile(const std::string &filename);
    /// LUA_EXPORT_OFF ///
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A short \\a sound that can be loaded from a file. Note: a sound should not be used to load musics ( > 1 min) since the whole file is loaded into memory. */
/// LUA_HEADER_COPY_OFF ///
class Sound : public sf::SoundBuffer, public ClassAllocCounter, public ProtectedClassBase
{
public:
    /** Destructor (will notice SoundPlayers) */
    virtual ~Sound();

    /// LUA_EXPORT_ON(Sound) ///
    /** Constructor to create an empty sound */
    Sound();

    /** Create a sound and load it from a file */
    Sound(const std::string &filename);

    /** Load a sound from a file. Note: the working directory is the mode's root directory. */
    void LoadFromFile(const std::string &filename);

	/** Get the sample rate
	  * \return Sound frequency (number of samples per second)
	  */
    int GetSampleRate() const;


	/** Return the number of channels (1 = mono, 2 = stereo, ...)
	  * \return Number of channels
	  */
    int GetChannelsCount() const;

	/** Get the sound duration
	  * \return Sound duration, in seconds
	  */
    //LUA// float GetDuration() const;
    /// LUA_EXPORT_OFF ///


    // No Lua
    //typedef std::pair<Screen*, sf::SoundBuffer*> sndBufferPtr;
    typedef SoundPlayer* sndBufferPtr;

    // SoundBuffer delete callback
    void AddSoundBufferCallback(sndBufferPtr ptr);
    void RemoveSoundBufferCallback(sndBufferPtr ptr);

protected:
    void MakeSoundBufferCallback();
    std::set<sndBufferPtr> m_callbackSet;
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A music. Note: a music is not attached to a screen. It can be played by itself. */
/// LUA_HEADER_COPY_OFF ///
class Music : public sf::Music, public ClassAllocCounter
{
public:
    /// LUA_EXPORT_ON(Music) ///
    /** Constructor to create an empty music */
    Music();

    /** Create a music and load it from a file */
    Music(const std::string &filename);

    /** Open a music file (doesn't play it -- call Play() for that)
      * Note: the working directory is the mode's root directory.
      * \\param filename : Path of the music file to open
      * \\return True if loading has been successful
      */
    void LoadFromFile(const std::string &filename);

    /** Get the music duration
      * \\return Music duration, in seconds
      */
    //LUA// float GetDuration() const;

    /** Return the number of channels (1 = mono, 2 = stereo, ...)
      * \\return Number of channels
      */
    //LUA// unsigned int GetChannelsCount() const;

    /** Play the sound
      */
    //LUA// void Play();

    /** Pause the sound
      */
    //LUA// void Pause();

    /** Stop the sound
      */
    //LUA// void Stop();

    /** Set the sound loop state.
      * This parameter is disabled by default
      * \\param Loop : True to play in loop, false to play once
      */
    //LUA// void SetLoop(bool Loop);

    /** Set the sound pitch.
      * The default pitch is 1
      * \\param Pitch : New pitch
      */
    //LUA// void SetPitch(float Pitch);

    /** Set the sound volume.
      * The default volume is 100
      * \\param Volume : Volume (in range [0, 100])
      */
    //LUA// void SetVolume(float Volume);

    // /** Set the current playing position of the sound
    //  * \\param TimeOffset : New playing position, expressed in seconds
    //  */
    // //LUA// void SetPlayingOffset(float TimeOffset);

    /** Tell whether or not the sound is looping
      * \\return True if the sound is looping, false otherwise
      */
    //LUA// bool GetLoop() const;

    /** Get the pitch
      * \\return Pitch value
      */
    //LUA// float GetPitch() const;

    /** Get the volume
      * \\return Volume value (in range [1, 100])
      */
    //LUA// float GetVolume() const;

    /** Get the current playing position of the sound
      * \\return Current playing position, expressed in seconds
      */
    //LUA// float GetPlayingOffset() const;

    /** Return true if the music is playing.
      * \\return Current status of the sound (playing or not)
      */
    bool IsPlaying() const;

    /// LUA_EXPORT_OFF ///
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A text font, that can be loaded from a file. */
/// LUA_HEADER_COPY_OFF ///
class Font : public sf::Font, public ClassAllocCounter, public ProtectedClassBase
{
public:
    /// LUA_EXPORT_ON(Font) ///
    /** Constructor to create an empty font */
    Font();

    /** Create a font and load it from a file */
    Font(const std::string &filename);

    /** Load a texture from a file. Note: the working directory is the mode's root directory. */
    void LoadFromFile(const std::string &filename);
    /// LUA_EXPORT_OFF ///
};


//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // BOTTLE_UTILS_HPP
